import json
import uuid
import logging

from kubernetes import client as k8s_client
from kubernetes.client.rest import ApiException

from kubeflow.fairing.constants import constants
from kubeflow.fairing.deployers.job.job import Job

logger = logging.getLogger(__name__)


class Serving(Job):
    """Serves a prediction endpoint using Kubernetes deployments and services"""

    def __init__(self, serving_class, namespace=None, runs=1, labels=None,
                 service_type="ClusterIP", pod_spec_mutators=None):
        """

        :param serving_class: the name of the class that holds the predict function.
        :param namespace: The k8s namespace where the it will be deployed.
        :param runs:
        :param labels: label for deployed service
        :param service_type: service type
        :param pod_spec_mutators: pod spec mutators (Default value = None)
        """
        super(Serving, self).__init__(namespace, runs,
                                      deployer_type=constants.SERVING_DEPLOPYER_TYPE,
                                      labels=labels)
        self.serving_class = serving_class
        self.service_type = service_type
        self.pod_spec_mutators = pod_spec_mutators or []

    def deploy(self, pod_spec):
        """deploy a seldon-core REST service

        :param pod_spec: pod spec for the service

        """
        self.job_id = str(uuid.uuid1())
        self.labels['fairing-id'] = self.job_id
        for fn in self.pod_spec_mutators:
            fn(self.backend, pod_spec, self.namespace)
        pod_template_spec = self.generate_pod_template_spec(pod_spec)
        pod_template_spec.spec.containers[0].command = ["seldon-core-microservice",
                                                        self.serving_class, "REST",
                                                        "--service-type=MODEL", "--persistence=0"]
        self.deployment_spec = self.generate_deployment_spec(pod_template_spec)
        self.service_spec = self.generate_service_spec()

        if self.output:
            api = k8s_client.ApiClient()
            job_output = api.sanitize_for_serialization(self.deployment_spec)
            logger.warning(json.dumps(job_output))
            service_output = api.sanitize_for_serialization(self.service_spec)
            logger.warning(json.dumps(service_output))

        v1_api = k8s_client.CoreV1Api()
        apps_v1 = k8s_client.AppsV1Api()
        self.deployment = apps_v1.create_namespaced_deployment(self.namespace, self.deployment_spec)
        self.service = v1_api.create_namespaced_service(self.namespace, self.service_spec)

        if self.service_type == "LoadBalancer":
            url = self.backend.get_service_external_endpoint(
                self.service.metadata.name, self.service.metadata.namespace,
                self.service.metadata.labels)
        else:
            # TODO(jlewi): The suffix won't always be cluster.local since
            # its configurable. Is there a way to get it programmatically?
            url = "http://{0}.{1}.svc.cluster.local:5000/predict".format(
                self.service.metadata.name, self.service.metadata.namespace)

        logging.info("Cluster endpoint: %s", url)
        return url

    def generate_deployment_spec(self, pod_template_spec):
        """generate deployment spec(V1Deployment)

        :param pod_template_spec: pod spec template

        """
        return k8s_client.V1Deployment(
            api_version="apps/v1",
            kind="Deployment",
            metadata=k8s_client.V1ObjectMeta(
                generate_name="fairing-deployer-",
                labels=self.labels,
            ),
            spec=k8s_client.V1DeploymentSpec(
                selector=k8s_client.V1LabelSelector(
                    match_labels=self.labels,
                ),
                template=pod_template_spec,
            )
        )

    def generate_service_spec(self):
        """ generate service spec(V1ServiceSpec)"""
        return k8s_client.V1Service(
            api_version="v1",
            kind="Service",
            metadata=k8s_client.V1ObjectMeta(
                generate_name="fairing-service-",
                labels=self.labels,
            ),
            spec=k8s_client.V1ServiceSpec(
                selector=self.labels,
                ports=[k8s_client.V1ServicePort(
                    name="serving",
                    port=5000
                )],
                type=self.service_type,
            )
        )

    def delete(self):
        """ delete the deployed service"""
        v1_api = k8s_client.CoreV1Api()
        try:
            v1_api.delete_namespaced_service(self.service.metadata.name, #pylint:disable=no-value-for-parameter
                                             self.service.metadata.namespace)
            logger.info("Deleted service: {}/{}".format(self.service.metadata.namespace,
                                                        self.service.metadata.name))
        except ApiException as e:
            logger.error(e)
            logger.error("Not able to delete service: {}/{}".format(self.service.metadata.namespace,
                                                                    self.service.metadata.name))
        try:
            api_instance = k8s_client.ExtensionsV1beta1Api()
            del_opts = k8s_client.V1DeleteOptions(propagation_policy="Foreground")
            api_instance.delete_namespaced_deployment(self.deployment.metadata.name,
                                                      self.deployment.metadata.namespace,
                                                      body=del_opts)
            logger.info("Deleted deployment: {}/{}".format(self.deployment.metadata.namespace,
                                                           self.deployment.metadata.name))
        except ApiException as e:
            logger.error(e)
            logger.error("Not able to delete deployment: {}/{}"\
                         .format(self.deployment.metadata.namespace, self.deployment.metadata.name))



import abc
import six


@six.add_metaclass(abc.ABCMeta)
class DeployerInterface(object):
    """Deploys a training job to the cluster"""

    def deploy(self, pod_template_spec):
        """Deploys the training job

        :param pod_template_spec: pod template spec

        """
        raise NotImplementedError('TrainingInterface.deploy')

    @abc.abstractmethod
    def get_logs(self):
        """Streams the logs for the training job"""
        raise NotImplementedError('TrainingInterface.train')



import uuid

from kubernetes import client

from kubeflow.fairing import utils
from kubeflow.fairing.builders.cluster.context_source import ContextSourceInterface
from kubeflow.fairing.cloud import azure
from kubeflow.fairing.constants import constants


class StorageContextSource(ContextSourceInterface):
    """Azure storage context source"""
    def __init__(self, namespace=None, region=None,
                 resource_group_name=None, storage_account_name=None):
        self.namespace = namespace or utils.get_default_target_namespace()
        self.region = region or "NorthEurope"
        self.resource_group_name = resource_group_name or "fairing"
        self.storage_account_name = storage_account_name or "fairing{}".format(
            uuid.uuid4().hex[:17]
        )
        self.share_name = constants.AZURE_FILES_SHARED_FOLDER
        self.context_hash = None
        self.context_path = None

    def prepare(self, context_filename):  # pylint:disable=arguments-differ
        self.context_hash = utils.crc(context_filename)
        self.context_path = self.upload_context(context_filename)

    def upload_context(self, context_filename):
        # Kaniko doesn't support Azure Storage yet.
        # So instead of uploading the context tar.gz file to Azure Storage
        # we are uploading the files in the context to a shared folder in Azure Files,
        # mounting the shared folder into the Kaniko pod,
        # and providing Kaniko with a local path to the files.
        azure_uploader = azure.AzureFileUploader(self.namespace)
        dir_name = "build_{}".format(self.context_hash)
        storage_account_name, storage_key = azure_uploader.upload_to_share(
            self.region,
            self.resource_group_name,
            self.storage_account_name,
            self.share_name,
            dir_name=dir_name,
            tar_gz_file_to_upload=context_filename)

        # This is the secret that we need to mount the shared folder into the Kaniko pod
        azure.create_storage_creds_secret(
            self.namespace, self.context_hash, storage_account_name, storage_key
        )

        # Local path to the files
        return "/mnt/azure/{}/".format(dir_name)


    def cleanup(self):
        azure.delete_storage_creds_secret(self.namespace, self.context_hash)

    def generate_pod_spec(self, image_name, push):  # pylint:disable=arguments-differ
        args = ["--dockerfile=Dockerfile",
                "--destination={}".format(image_name),
                "--context={}".format(self.context_path)]
        if not push:
            args.append("--no-push")
        return client.V1PodSpec(
            containers=[client.V1Container(
                name='kaniko',
                image=constants.KANIKO_IMAGE,
                args=args,
            )],
            restart_policy='Never'
        )



import abc
import six
import sys
import logging

from kubeflow.fairing import utils
from kubeflow.fairing.builders.docker.docker import DockerBuilder
from kubeflow.fairing.builders.cluster import gcs_context
from kubeflow.fairing.builders.cluster.cluster import ClusterBuilder
from kubeflow.fairing.builders.cluster import s3_context
from kubeflow.fairing.builders.cluster import azurestorage_context
from kubeflow.fairing.builders.append.append import AppendBuilder
from kubeflow.fairing.deployers.gcp.gcp import GCPJob
from kubeflow.fairing.deployers.job.job import Job
from kubeflow.fairing.deployers.serving.serving import Serving
from kubeflow.fairing.cloud import aws
from kubeflow.fairing.cloud import azure
from kubeflow.fairing.cloud import gcp
from kubeflow.fairing.cloud import docker
from kubeflow.fairing.ml_tasks import utils as ml_tasks_utils
from kubeflow.fairing.constants import constants
from kubeflow.fairing.kubernetes.manager import KubeManager

logger = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class BackendInterface(object):
    """ Backend interface.
    Creating a builder instance or a deployer to be used with a traing job or a serving job
    for the given backend.
    And get the approriate base container or docker registry for the current environment.
    """

    @abc.abstractmethod
    def get_builder(self, preprocessor, base_image, registry):
        """Creates a builder instance with right config for the given backend

        :param preprocessor: Preprocessor to use to modify inputs
        :param base_image: Base image to use for this builder
        :param registry: Registry to push image to. Example: gcr.io/kubeflow-images
        :raises NotImplementedError: not implemented exception

        """

        raise NotImplementedError('BackendInterface.get_builder')

    @abc.abstractmethod
    def get_training_deployer(self, pod_spec_mutators=None):
        """Creates a deployer to be used with a training job

        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
                                  e.g. fairing.cloud.gcp.add_gcp_credentials_if_exists
                                  This can used to set things like volumes and security context.
                                  (Default value = None)
        :raises NotImplementedError: not implemented exception

        """

        raise NotImplementedError('BackendInterface.get_training_deployer')

    @abc.abstractmethod
    def get_serving_deployer(self, model_class):
        """Creates a deployer to be used with a serving job

        :param model_class: the name of the class that holds the predict function.
        :raises NotImplementedError: not implemented exception

        """

        raise NotImplementedError('BackendInterface.get_serving_deployer')

    def get_base_contanier(self):
        """Returns the approriate base container for the current environment

        :returns: base image

        """

        py_version = ".".join([str(x) for x in sys.version_info[0:3]])
        base_image = 'registry.hub.docker.com/library/python:{}'.format(
            py_version)
        return base_image

    def get_docker_registry(self):
        """Returns the approriate docker registry for the current environment

        :returns: None

        """

        return None


class KubernetesBackend(BackendInterface):
    """ Use to create a builder instance and create a deployer to be used with a traing job or
    a serving job for the Kubernetes.
    """
    def __init__(self, namespace=None, build_context_source=None):
        if not namespace and not utils.is_running_in_k8s():
            logger.warning("Can't determine namespace automatically. "
                           "Using 'default' namespace but recomend to provide namespace explicitly"
                           ". Using 'default' namespace might result in unable to mount some "
                           "required secrets in cloud backends.")
        self._namespace = namespace or utils.get_default_target_namespace()
        self._build_context_source = build_context_source

    def get_builder(self, preprocessor, base_image, registry, needs_deps_installation=True,  # pylint:disable=arguments-differ
                    pod_spec_mutators=None):
        """Creates a builder instance with right config for the given Kubernetes

        :param preprocessor: Preprocessor to use to modify inputs
        :param base_image: Base image to use for this job
        :param registry: Registry to push image to. Example: gcr.io/kubeflow-images
        :param needs_deps_installation:  need depends on installation(Default value = True)
        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
                                  e.g. fairing.cloud.gcp.add_gcp_credentials_if_exists
                                  This can used to set things like volumes and security context.
                                  (Default value =None)

        """
        if not needs_deps_installation:
            return AppendBuilder(preprocessor=preprocessor,
                                 base_image=base_image,
                                 registry=registry)
        elif utils.is_running_in_k8s():
            return ClusterBuilder(preprocessor=preprocessor,
                                  base_image=base_image,
                                  registry=registry,
                                  pod_spec_mutators=pod_spec_mutators,
                                  namespace=self._namespace,
                                  context_source=self._build_context_source)
        elif ml_tasks_utils.is_docker_daemon_exists():
            return DockerBuilder(preprocessor=preprocessor,
                                 base_image=base_image,
                                 registry=registry)
        else:
            # TODO (karthikv2k): Add more info on how to reolve this issue
            raise RuntimeError(
                "Not able to guess the right builder for this job!")

    def get_training_deployer(self, pod_spec_mutators=None):
        """Creates a deployer to be used with a training job for the Kubernetes

        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
            (Default value = None)
        :returns: job for handle all the k8s' template building for a training

        """
        return Job(self._namespace, pod_spec_mutators=pod_spec_mutators)

    def get_serving_deployer(self, model_class, service_type='ClusterIP', # pylint:disable=arguments-differ
                             pod_spec_mutators=None):
        """Creates a deployer to be used with a serving job for the Kubernetes

        :param model_class: the name of the class that holds the predict function.
        :param service_type: service type (Default value = 'ClusterIP')
        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
            (Default value = None)

        """
        return Serving(model_class, namespace=self._namespace, service_type=service_type,
                       pod_spec_mutators=pod_spec_mutators)


class GKEBackend(KubernetesBackend):
    """ Use to create a builder instance and create a deployer to be used with a traing job
    or a serving job for the GKE backend.
    And get the approriate docker registry for GKE.
    """

    def __init__(self, namespace=None, build_context_source=None):
        super(GKEBackend, self).__init__(namespace, build_context_source)
        self._build_context_source = gcs_context.GCSContextSource(
            namespace=self._namespace)

    def get_builder(self, preprocessor, base_image, registry, needs_deps_installation=True,
                    pod_spec_mutators=None):
        """Creates a builder instance with right config for GKE

        :param preprocessor: Preprocessor to use to modify inputs
        :param base_image: Base image to use for this job
        :param registry: Registry to push image to. Example: gcr.io/kubeflow-images
        :param needs_deps_installation:  need depends on installation(Default value = True)
        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
                                  e.g. fairing.cloud.gcp.add_gcp_credentials_if_exists
                                  This can used to set things like volumes and security context.
                                  (Default value =None)

        """

        pod_spec_mutators = pod_spec_mutators or []
        pod_spec_mutators.append(gcp.add_gcp_credentials_if_exists)

        if not needs_deps_installation:
            return AppendBuilder(preprocessor=preprocessor,
                                 base_image=base_image,
                                 registry=registry)
        elif (utils.is_running_in_k8s() or
              not ml_tasks_utils.is_docker_daemon_exists()):
            return ClusterBuilder(preprocessor=preprocessor,
                                  base_image=base_image,
                                  registry=registry,
                                  pod_spec_mutators=pod_spec_mutators,
                                  namespace=self._namespace,
                                  context_source=self._build_context_source)
        elif ml_tasks_utils.is_docker_daemon_exists():
            return DockerBuilder(preprocessor=preprocessor,
                                 base_image=base_image,
                                 registry=registry)
        else:
            msg = ["Not able to guess the right builder for this job!"]
            if KubeManager().secret_exists(constants.GCP_CREDS_SECRET_NAME, self._namespace):
                msg.append("It seems you don't have permission to list/access secrets in your "
                           "Kubeflow cluster. We need this permission in order to build a docker "
                           "image using Kubeflow cluster. Adding Kubeneters Admin role to the "
                           "service account you are using might solve this issue.")
            if not utils.is_running_in_k8s():
                msg.append(" Also If you are using 'sudo' to access docker in your system you can"
                           " solve this problem by adding your username to the docker group. "
                           "Reference: https://docs.docker.com/install/linux/linux-postinstall/"
                           "#manage-docker-as-a-non-root-user You need to logout and login to "
                           "get change activated.")
            message = " ".join(msg)
            raise RuntimeError(message)

    def get_training_deployer(self, pod_spec_mutators=None):
        """Creates a deployer to be used with a training job for GKE

        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
            (Default value = None)
        :returns: job for handle all the k8s' template building for a training

        """
        pod_spec_mutators = pod_spec_mutators or []
        pod_spec_mutators.append(gcp.add_gcp_credentials_if_exists)
        return Job(namespace=self._namespace, pod_spec_mutators=pod_spec_mutators)

    def get_serving_deployer(self, model_class, service_type='ClusterIP',
                             pod_spec_mutators=None):
        """Creates a deployer to be used with a serving job for GKE

        :param model_class: the name of the class that holds the predict function.
        :param service_type: service type (Default value = 'ClusterIP')
        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
            (Default value = None)

        """
        return Serving(model_class, namespace=self._namespace, service_type=service_type,
                       pod_spec_mutators=pod_spec_mutators)

    def get_docker_registry(self):
        """Returns the approriate docker registry for GKE

        :returns: docker registry

        """
        return gcp.get_default_docker_registry()


class AWSBackend(KubernetesBackend):
    """ Use to create a builder instance and create a deployer to be used with a traing job
    or a serving job for the AWS backend.
    """

    def __init__(self, namespace=None, build_context_source=None):
        build_context_source = build_context_source or s3_context.S3ContextSource()
        super(AWSBackend, self).__init__(namespace, build_context_source)

    def get_builder(self, preprocessor, base_image, registry, needs_deps_installation=True,
                    pod_spec_mutators=None):
        """Creates a builder instance with right config for AWS

        :param preprocessor: Preprocessor to use to modify inputs
        :param base_image: Base image to use for this job
        :param registry: Registry to push image to. Example: gcr.io/kubeflow-images
        :param needs_deps_installation:  need depends on installation(Default value = True)
        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
                                  e.g. fairing.cloud.gcp.add_gcp_credentials_if_exists
                                  This can used to set things like volumes and security context.
                                  (Default value =None)

        """
        pod_spec_mutators = pod_spec_mutators or []
        pod_spec_mutators.append(aws.add_aws_credentials_if_exists)
        if aws.is_ecr_registry(registry):
            pod_spec_mutators.append(aws.add_ecr_config)
            aws.create_ecr_registry(registry, constants.DEFAULT_IMAGE_NAME)
        return super(AWSBackend, self).get_builder(preprocessor,
                                                   base_image,
                                                   registry,
                                                   needs_deps_installation,
                                                   pod_spec_mutators)

    def get_training_deployer(self, pod_spec_mutators=None):
        """Creates a deployer to be used with a training job for AWS

        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
            (Default value = None)
        :returns: job for handle all the k8s' template building for a training

        """
        pod_spec_mutators = pod_spec_mutators or []
        pod_spec_mutators.append(aws.add_aws_credentials_if_exists)
        return Job(namespace=self._namespace, pod_spec_mutators=pod_spec_mutators)

    def get_serving_deployer(self, model_class, service_type='ClusterIP', # pylint:disable=arguments-differ
                             pod_spec_mutators=None):
        """Creates a deployer to be used with a serving job for AWS

        :param model_class: the name of the class that holds the predict function.
        :param service_type: service type (Default value = 'ClusterIP')
        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
            (Default value = None)

        """
        return Serving(model_class, namespace=self._namespace, service_type=service_type,
                       pod_spec_mutators=pod_spec_mutators)

class AzureBackend(KubernetesBackend):
    """ Use to create a builder instance and create a deployer to be used with a traing job or
    a serving job for the Azure backend.
    """
    def __init__(self, namespace=None, build_context_source=None):
        build_context_source = (
            build_context_source or azurestorage_context.StorageContextSource(namespace=namespace)
        )
        super(AzureBackend, self).__init__(namespace, build_context_source)

    def get_builder(self, preprocessor, base_image, registry,
                    needs_deps_installation=True, pod_spec_mutators=None):
        """Creates a builder instance with right config for Azure

        :param preprocessor: Preprocessor to use to modify inputs
        :param base_image: Base image to use for this job
        :param registry: Registry to push image to. Example: gcr.io/kubeflow-images
        :param needs_deps_installation:  need depends on installation(Default value = True)
        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
                                  e.g. fairing.cloud.gcp.add_gcp_credentials_if_exists
                                  This can used to set things like volumes and security context.
                                  (Default value =None)

        """
        pod_spec_mutators = pod_spec_mutators or []
        if not azure.is_acr_registry(registry):
            raise Exception("'{}' is not an Azure Container Registry".format(registry))
        pod_spec_mutators.append(azure.add_acr_config)
        pod_spec_mutators.append(azure.add_azure_files)
        return super(AzureBackend, self).get_builder(preprocessor,
                                                     base_image,
                                                     registry,
                                                     needs_deps_installation,
                                                     pod_spec_mutators)

class KubeflowBackend(KubernetesBackend):
    """Kubeflow backend refer to KubernetesBackend """

    def __init__(self, namespace=None, build_context_source=None):
        if not namespace and not utils.is_running_in_k8s():
            namespace = "kubeflow"
        super(KubeflowBackend, self).__init__(namespace, build_context_source)


class KubeflowGKEBackend(GKEBackend):
    """Kubeflow for GKE backend refer to GKEBackend """

    def __init__(self, namespace=None, build_context_source=None):
        if not namespace and not utils.is_running_in_k8s():
            namespace = "kubeflow"
        super(KubeflowGKEBackend, self).__init__(
            namespace, build_context_source)


class KubeflowAWSBackend(AWSBackend):
    """Kubeflow for AWS backend refer to AWSBackend """

    def __init__(self, namespace=None, build_context_source=None): # pylint:disable=useless-super-delegation
        super(KubeflowAWSBackend, self).__init__(
            namespace, build_context_source)


class KubeflowAzureBackend(AzureBackend):
    """Kubeflow for Azure backend refer to AzureBackend """

    def __init__(self, namespace=None, build_context_source=None): # pylint:disable=useless-super-delegation
        super(KubeflowAzureBackend, self).__init__(namespace, build_context_source)


class GCPManagedBackend(BackendInterface):
    """ Use to create a builder instance and create a deployer to be used with a traing job
    or a serving job for the GCP.
    """
    def __init__(self, project_id=None, region=None, training_scale_tier=None):
        super(GCPManagedBackend, self).__init__()
        self._project_id = project_id or gcp.guess_project_name()
        self._region = region or 'us-central1'
        self._training_scale_tier = training_scale_tier or 'BASIC'

    def get_builder(self, preprocessor, base_image, registry, needs_deps_installation=True,  # pylint:disable=arguments-differ
                    pod_spec_mutators=None):
        """Creates a builder instance with right config for GCP

        :param preprocessor: Preprocessor to use to modify inputs
        :param base_image: Base image to use for this job
        :param registry: Registry to push image to. Example: gcr.io/kubeflow-images
        :param needs_deps_installation:  need depends on installation(Default value = True)
        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
                                  e.g. fairing.cloud.gcp.add_gcp_credentials_if_exists
                                  This can used to set things like volumes and security context.
                                  (Default value =None)

        """
        pod_spec_mutators = pod_spec_mutators or []
        pod_spec_mutators.append(gcp.add_gcp_credentials_if_exists)
        pod_spec_mutators.append(docker.add_docker_credentials_if_exists)
        # TODO (karthikv2k): Add cloud build as the deafult
        # once https://github.com/kubeflow/fairing/issues/145 is fixed
        if not needs_deps_installation:
            return AppendBuilder(preprocessor=preprocessor,
                                 base_image=base_image,
                                 registry=registry)
        elif utils.is_running_in_k8s():
            return ClusterBuilder(preprocessor=preprocessor,
                                  base_image=base_image,
                                  registry=registry,
                                  pod_spec_mutators=pod_spec_mutators,
                                  context_source=gcs_context.GCSContextSource(
                                      namespace=utils.get_default_target_namespace()))
        elif ml_tasks_utils.is_docker_daemon_exists():
            return DockerBuilder(preprocessor=preprocessor,
                                 base_image=base_image,
                                 registry=registry)
        else:
            # TODO (karthikv2k): Add more info on how to reolve this issue
            raise RuntimeError(
                "Not able to guess the right builder for this job!")

    def get_training_deployer(self, pod_spec_mutators=None):
        """Creates a deployer to be used with a training job for GCP

        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
            (Default value = None)
        :returns: job for handle all the k8s' template building for a training

        """
        return GCPJob(self._project_id, self._region, self._training_scale_tier)

    def get_serving_deployer(self, model_class, pod_spec_mutators=None): # pylint:disable=arguments-differ
        """Creates a deployer to be used with a serving job for GCP

        :param model_class: the name of the class that holds the predict function.
        :param service_type: service type (Default value = 'ClusterIP')
        :param pod_spec_mutators: list of functions that is used to mutate the podsspec.
            (Default value = None)

        """
        # currently GCP serving deployer doesn't implement deployer interface
        raise NotImplementedError(
            "GCP managed serving is not integrated into high level API yet.")

    def get_docker_registry(self):
        """Returns the approriate docker registry for GCP

        :returns: docker registry

        """
        return gcp.get_default_docker_registry()



import os

TEMP_TAR_GZ_FILENAME = '/tmp/fairing.layer.tar.gz'
DEFAULT_IMAGE_NAME = 'fairing-job'
DEFAULT_BASE_IMAGE = 'gcr.io/kubeflow-images-public/fairing:dev'
DEFAULT_REGISTRY = 'index.docker.io'
DEFAULT_DEST_PREFIX = '/app/'

DEFAULT_CONTEXT_FILENAME = '/tmp/fairing.context.tar.gz'
DEFAULT_GENERATED_DOCKERFILE_FILENAME = '/tmp/Dockerfile'

GOOGLE_CREDS_ENV = 'GOOGLE_APPLICATION_CREDENTIALS'
GCP_SERVICE_ACCOUNT_NAME = 'default-editor'

GCP_CREDS_SECRET_NAME = 'user-gcp-sa'
AWS_CREDS_SECRET_NAME = 'aws-secret'
DOCKER_CREDS_SECRET_NAME = "docker-secret"

# See https://github.com/kubeflow/website/issues/1033 for documentation about these secrets.
AZURE_CREDS_SECRET_NAME = 'azcreds'
AZURE_ACR_CREDS_SECRET_NAME = 'acrcreds'
# The secret containing credentials to access a specific storage account is dynamically generated
# by using Azure credentials to get those storage credentials.
AZURE_STORAGE_CREDS_SECRET_NAME_PREFIX = 'storage-credentials-'
AZURE_FILES_SHARED_FOLDER = 'fairing-builds'

DEFAULT_USER_AGENT = 'kubeflow-fairing/{VERSION}'

# Job Constants
JOB_DEFAULT_NAME = 'fairing-job-'
JOB_DEPLOPYER_TYPE = 'job'

# Serving Constants
SERVING_DEPLOPYER_TYPE = 'serving'

#TFJob Constants
TF_JOB_VERSION = os.environ.get('TF_JOB_VERSION', 'v1')
TF_JOB_GROUP = "kubeflow.org"
TF_JOB_KIND = "TFJob"
TF_JOB_PLURAL = "tfjobs"
TF_JOB_DEFAULT_NAME = 'fairing-tfjob-'
TF_JOB_DEPLOYER_TYPE = 'tfjob'

# KFServing constants
KFSERVING_GROUP = "serving.kubeflow.org"
KFSERVING_KIND = "InferenceService"
KFSERVING_PLURAL = "inferenceservices"
KFSERVING_VERSION = 'v1alpha2'
KFSERVING_DEFAULT_NAME = 'fairing-kfserving-'
KFSERVING_DEPLOYER_TYPE = 'kfservice'
KFSERVING_CONTAINER_NAME = 'user-container'

# persistent volume claim constants
PVC_DEFAULT_MOUNT_PATH = '/mnt'
PVC_DEFAULT_VOLUME_NAME = 'fairing-volume-'

# Kaniko Constants
KANIKO_IMAGE = 'gcr.io/kaniko-project/executor:v0.7.0'



import logging
import json
import numpy as np
from kubeflow.fairing.backends import KubernetesBackend
from kubeflow.fairing.ml_tasks.utils import guess_preprocessor

import requests

logger = logging.getLogger(__name__)


class BaseTask:
    """
    Base class for handling high level ML tasks.

    :param entry_point: An object or reference to the source code that has to be deployed.
    :param base_docker_image: Name of the base docker image that should be used as a base image
           when building a new docker image as part of an ML task deployment.
    :param docker_registry: Docker registry to store output docker images.
    :param input_files: list of files that needs to be packaged along with the entry point.
           E.g. local python modules, trained model weigths, etc.
    """

    def __init__(self, entry_point, base_docker_image=None, docker_registry=None,
                 input_files=None, backend=None, pod_spec_mutators=None):
        self._backend = backend or KubernetesBackend()
        self._pod_spec_mutators = pod_spec_mutators or []
        input_files = input_files or []
        output_map = {}

        preprocessor = guess_preprocessor(entry_point,
                                          input_files=input_files,
                                          output_map=output_map)
        logger.debug("Using preprocessor: {}".format(type(preprocessor)))

        self.docker_registry = docker_registry or backend.get_docker_registry()
        if not self.docker_registry:
            raise RuntimeError("Not able find a default docker registry."
                               " Please provide 'docker_registry' argument explicitly."
                               " Docker registry is used to store the output docker images"
                               " that are executed in the remote cluster.")
        if not docker_registry:
            logger.warning("Using default docker registry: {}".format(
                self.docker_registry))

        self.base_docker_image = base_docker_image or backend.get_base_contanier()
        if not self.base_docker_image:
            raise RuntimeError("Not able find a default base docker image."
                               " Please provide 'base_docker_image' argument explicitly."
                               " Base docker image is used to build the output docker images"
                               " that are executed in the remote cluster.")
        if not base_docker_image:
            logger.warning("Using default base docker image: {}".format(
                self.base_docker_image))

        needs_deps_installation = "requirements.txt" in input_files
        self.builder = self._backend.get_builder(preprocessor=preprocessor,
                                                 base_image=self.base_docker_image,
                                                 registry=self.docker_registry,
                                                 pod_spec_mutators=self._pod_spec_mutators,
                                                 needs_deps_installation=needs_deps_installation)
        logger.warning("Using builder: {}".format(type(self.builder)))

    def _build(self):
        """Build the docker image. """
        logging.info("Building the docker image.")
        self.builder.build()
        self.pod_spec = self.builder.generate_pod_spec()


class TrainJob(BaseTask):
    """Create a train job. """

    def __init__(self, entry_point, base_docker_image=None, docker_registry=None,  # pylint:disable=useless-super-delegation
                 input_files=None, backend=None, pod_spec_mutators=None):
        super().__init__(entry_point, base_docker_image, docker_registry,
                         input_files, backend, pod_spec_mutators)

    def submit(self):
        """Submit a train job. """
        self._build()
        deployer = self._backend.get_training_deployer(
            pod_spec_mutators=self._pod_spec_mutators)
        return deployer.deploy(self.pod_spec)


class PredictionEndpoint(BaseTask):
    """Create a prediction endpoint. """

    def __init__(self, model_class, base_docker_image=None, docker_registry=None, input_files=None,
                 backend=None, service_type='ClusterIP', pod_spec_mutators=None):
        self.model_class = model_class
        self.service_type = service_type
        super().__init__(model_class, base_docker_image, docker_registry,
                         input_files, backend, pod_spec_mutators)

    def create(self):
        """Create prediction endpoint. """
        self._build()
        logging.info("Deploying the endpoint.")
        self._deployer = self._backend.get_serving_deployer(
            self.model_class.__name__,
            service_type=self.service_type,
            pod_spec_mutators=self._pod_spec_mutators)
        self.url = self._deployer.deploy(self.pod_spec)
        logger.warning("Prediction endpoint: {}".format(self.url))

    def predict_nparray(self, data, feature_names=None):
        """Return the prediction result.

        :param data: Data to be predicted.
        :param feature_names: Feature extracted from data (Default value = None)

        """
        pdata = {
            "data": {
                "names": feature_names,
                "tensor": {
                    "shape": np.asarray(data.shape).tolist(),
                    "values": data.flatten().tolist(),
                },
            }
        }
        serialized_data = json.dumps(pdata)
        r = requests.post(self.url, data={'json': serialized_data})
        return json.loads(r.text)

    def delete(self):
        """Delete prediction endpoint. """
        logging.info("Deleting the endpoint. ")
        self._deployer.delete()



import logging
from kubernetes import client as k8s_client

from kubeflow.fairing.constants import constants
from kubeflow.fairing.deployers.job.job import Job

logger = logging.getLogger(__name__)


class TfJob(Job):
    """ Handle all the k8s' template building to create tensorflow
        training job using Kubeflow TFOperator"""
    def __init__(self, namespace=None, worker_count=1, ps_count=0,
                 chief_count=1, runs=1, job_name=constants.TF_JOB_DEFAULT_NAME, stream_log=True,
                 labels=None, pod_spec_mutators=None, cleanup=False, annotations=None):
        """

        :param namespace: k8s namespace where the training's components
                will be deployed.
        :param worker_count: Number of worker pods created for training job.
        :param ps_count: Number of parameter set pods created for training job.
        :param chief_count: Number of Chief pods created for training job.
        :param runs: number of runs
        :param job_name: name of the job
        :param stream_log: stream the log from deployed tfjob
        :param labels: labels to be assigned to the training job
        :param pod_spec_mutators: pod spec mutators (Default value = None)
        :param cleanup: clean up deletes components after job finished
        :param annotations: annotations (Default value = None)
        """
        super(TfJob, self).__init__(namespace, runs, job_name=job_name, stream_log=stream_log,
                                    deployer_type=constants.TF_JOB_DEPLOYER_TYPE, labels=labels,
                                    pod_spec_mutators=pod_spec_mutators, cleanup=cleanup,
                                    annotations=annotations)
        self.distribution = {
            'Worker': worker_count,
            'PS': ps_count,
            'Chief': chief_count
        }

    def create_resource(self):
        """ create a tfjob training"""
        self.created_tfjob = self.backend.create_tf_job(self.namespace, self.deployment_spec)
        return self.created_tfjob['metadata']['name']

    def generate_deployment_spec(self, pod_template_spec):
        """Returns a TFJob template

        :param pod_template_spec: template spec for pod

        """
        self.set_container_name(pod_template_spec)

        worker_replica_spec = {}
        worker_replica_spec['replicas'] = self.distribution['Worker']
        worker_replica_spec['template'] = pod_template_spec

        ps_replica_spec = {}
        ps_replica_spec['replicas'] = self.distribution.get('PS', 0)
        ps_replica_spec['template'] = pod_template_spec

        chief_replica_spec = {}
        chief_replica_spec['replicas'] = self.distribution.get('Chief', 0)
        chief_replica_spec['template'] = pod_template_spec

        spec = {}
        spec['tfReplicaSpecs'] = {}
        spec['tfReplicaSpecs']['Worker'] = worker_replica_spec
        if chief_replica_spec['replicas'] > 0:
            spec['tfReplicaSpecs']['Chief'] = chief_replica_spec
        if ps_replica_spec['replicas'] > 0:
            spec['tfReplicaSpecs']['PS'] = ps_replica_spec

        tf_job = {}
        tf_job['kind'] = constants.TF_JOB_KIND
        tf_job['apiVersion'] = 'kubeflow.org/' + constants.TF_JOB_VERSION
        tf_job['metadata'] = k8s_client.V1ObjectMeta(generate_name=self.job_name,
                                                     labels=self.labels)
        tf_job['spec'] = spec

        return tf_job

    def set_container_name(self, pod_template_spec):
        """Sets the name of the main container to `tensorflow`.
            This is required for TfJobs

        :param pod_template_spec: spec for pod template

        """
        pod_template_spec.spec.containers[0].name = 'tensorflow'

    def get_logs(self):
        """ get logs"""
        name = self.created_tfjob['metadata']['name']
        namespace = self.created_tfjob['metadata']['namespace']

        labels = {
            'tf-replica-index': '0',
            'tf-replica-type': 'worker',
            'tf-job-name': name
        }
        self.backend.log(name, namespace, labels, container="tensorflow")

        if self.cleanup:
            logger.warning("Cleaning up TFJob {}...".format(name))
            self.backend.delete_tf_job(name, self.namespace)



import argparse
import cloudpickle
import sys
import types
from enum import Enum
import logging

logging.basicConfig(format='%(message)s')
logging.getLogger().setLevel(logging.INFO)

class ObjectType(Enum):
    FUNCTION = 1
    CLASS = 2
    NOT_SUPPORTED = 3

def get_execution_obj_type(obj):
    """Get the execution object type, the object can be a function or class.

    :param obj: The name of object such as the function or class
    :returns: int: The corresponding object type.

    """
    # Check if a function is provided
    if (isinstance(obj, types.FunctionType) or  #pylint:disable=consider-merging-isinstance
            isinstance(obj, types.BuiltinFunctionType) or
            isinstance(obj, types.BuiltinMethodType)):
        return ObjectType.FUNCTION

    # Reject class methods provided directly
    if isinstance(obj, types.MethodType):
        return ObjectType.NOT_SUPPORTED

    # Check if a class is provided (Python 3)
    if isinstance(obj, type) and hasattr(obj, 'train'):
        return ObjectType.CLASS

    return ObjectType.NOT_SUPPORTED


def compare_version(local_python_version):
    """Compare the Python major and minor version for local and remote python.

    :param local_python_version: Python version of local environment
    :returns: None.

    """
    remote_python_version = ".".join([str(x) for x in sys.version_info[0:2]])
    if local_python_version != remote_python_version:
        raise RuntimeError('The Python version ' + remote_python_version + ' mismatches \
                           with Python ' + local_python_version + ' in the local environment.')


def call(serialized_fn_file):
    """Get the content from serialized function.

    :param serialized_fn_file: the file includes serialized function
    :returns: object: The content of object.

    """
    with open(serialized_fn_file, 'rb') as f:
        obj = cloudpickle.load(f)
        obj_type = get_execution_obj_type(obj)
        if obj_type == ObjectType.FUNCTION:
            res = obj()
        elif obj_type == ObjectType.CLASS:
            res = obj().train()
        else:
            raise RuntimeError("Object must of type function or a \
                               class but got {}".format(type(obj)))
        if res:
            print(res)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Shim for calling a serialized function')
    parser.add_argument('--serialized_fn_file', action="store")
    parser.add_argument('--python_version', action="store")
    args = parser.parse_args()
    compare_version(args.python_version)
    call(args.serialized_fn_file)



import logging
from kubernetes import client
from docker.utils.config import find_config_file
from base64 import b64encode

from kubeflow.fairing.constants import constants

logger = logging.getLogger(__name__)


def get_docker_secret():
    try:
        docker_config_file = find_config_file(config_path=None)
        with open(docker_config_file, 'r') as f:
            data = f.read()
            data = {".dockerconfigjson": b64encode(
                data.encode('utf-8')).decode("utf-8")}
        docker_secret = client.V1Secret(
            metadata=client.V1ObjectMeta(name=constants.DOCKER_CREDS_SECRET_NAME),
            data=data,
            kind="Secret",
            type="kubernetes.io/dockerconfigjson"
        )
        return docker_secret
    except Exception as e:
        logger.warning("could not get docker secret: {}".format(e))
    return None


def create_docker_secret(kube_manager, namespace):
    try:
        docker_secret = get_docker_secret()
        if docker_secret:
            kube_manager.create_secret(namespace, docker_secret)
    except Exception as e:
        logger.warning("could not create docker secret: {}".format(e))


def add_docker_credentials_if_exists(kube_manager, pod_spec, namespace):
    secret_name = constants.DOCKER_CREDS_SECRET_NAME
    try:
        if not kube_manager.secret_exists(secret_name, namespace):
            create_docker_secret(kube_manager, namespace)
        if kube_manager.secret_exists(secret_name, namespace):
            add_docker_credentials(kube_manager, pod_spec, namespace)
        else:
            logger.warning("Not able to find docker credentials secret: {}".format(secret_name))
    except Exception as e:
        logger.warning("could not check for secret: {}".format(e))


def add_docker_credentials(kube_manager, pod_spec, namespace):
    secret_name = constants.DOCKER_CREDS_SECRET_NAME
    if not kube_manager.secret_exists(secret_name, namespace):
        raise ValueError("Not able to find docker credentials secret: {}".format(secret_name))
    pod_spec.image_pull_secrets = [client.V1LocalObjectReference(secret_name)]
